#!/bin/bash
#
# Copyright (C) 2015-2016 Trond Hasle Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Setting path
PATH=/bin:/usr/bin
export PATH

#---------------------------------------------------------------------
# Functions
#---------------------------------------------------------------------

# Cleanup function
function cleanup {
    if [ -n "$stdin_file" -a -f "$stdin_file" ]; then
        rm -f $stdin_file
    fi
}

#---------------------------------------------------------------------
# Variables and initialization
#---------------------------------------------------------------------

# Myself
ME=$(basename $0)

# Default options
PRINT_USER=${USER}
WORKGROUP='UIO'
SERVER='pushprint.uio.no'
PRINTER=${PRINTER}
PULLPRINT_QUEUE=${PULLPRINT_QUEUE:-ricoh}
COMMAND='print -'
KERBEROS='yes'
DRYRUN='no'

# Version and author info
VERSION='0.8'
AUTHOR='Trond Hasle Amundsen <t.h.amundsen@usit.uio.no>'
LICENSE=$'Copyright (C) 2015-2016 Trond Hasle Amundsen
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.'

# Usage and help text
USAGE_PUSH="Usage: pushprint [OPTION...] <file(s)>"
USAGE_PULL="Usage: pullprint [OPTION...] <file(s)>"
HELP_PUSH=$'
OPTIONS:
  -P <printer>  Printer name (name of Windows printer share)
  -S <server>   Server name (default: pushprint.uio.no)
  -W <domain>   Domain/Workgroup (default: UIO)
  -U <user>     Specify username. Not with Kerberos authentication
  -A <file>     Authentication file. Implies "-N"
  -N            Turn off Kerberos authentication (dangerous!)
  -E <encoding> Use encoding for converting text, if not recognized
  -n            Dryrun, do not actually print. For debugging
  -V            Display version info
  -h            This help text

Usually, only the "-P" option is necessary, to specify printer
queue. This option can be omitted if the shell variable $PRINTER
is set and contains a valid printer name.

This script will use an existing Kerberos ticket to authenticate, or
otherwise create a new ticket for authentication. Do not turn off
Kerberos authentication (option "-N") unless you know exactly what 
you are doing.'
HELP_PULL=$'
OPTIONS:
  -P <queue>    Pullprint queue name, "ricoh" (default) or "hp"
  -U <user>     Specify username
  -E <encoding> Use encoding for converting text, if not recognized
  -n            Dryrun, do not actually print. For debugging
  -V            Display version info
  -h            This help text

How the program chooses which printer queue to use:

  1. The default queue, which is "ricoh"; or
  2. The shell variable PULLPRINT_QUEUE, if defined; or
  3. The option "-P <queue>", if specified
'

# Read from STDIN if user is piping to this script
if read -t 0; then
    stdin_file=$(mktemp)
    cat - > $stdin_file
fi
exec </dev/tty >/dev/tty

# Getting options
while getopts ':x:hU:P:S:A:E:NVn' opt; do
    case $opt in
	x)
	    METHOD=${OPTARG}
	    ;;
        h)
	    if [ "$METHOD" == "pull" ]; then
		echo $USAGE_PULL
		echo "$HELP_PULL"
	    elif [ "$METHOD" == "push" ]; then
		echo $USAGE_PUSH
		echo "$HELP_PUSH"
	    else
		echo >&2 "ERROR: Wrong printing method: '$METHOD'. This shouldn't happen."
		cleanup
		exit 1
	    fi
            cleanup
            exit 0
            ;;
        U)
            PRINT_USER=${OPTARG}
            ;;
        P)
	    if [ "$METHOD" == "pull" ]; then
		PULLPRINT_QUEUE=${OPTARG}
	    elif [ "$METHOD" == "push" ]; then
		PRINTER=${OPTARG}
	    else
		echo >&2 "ERROR: Wrong printing method: '$METHOD'. This shouldn't happen."
		cleanup
		exit 1
	    fi		
            ;;
        S)
            SERVER=${OPTARG}
            ;;
        W)
            WORKGROUP=${OPTARG}
            ;;
        N)
            KERBEROS='no'
            ;;
	A)
	    AUTH_FILE=${OPTARG}
            KERBEROS='no'
	    ;;
        n)
	    DRYRUN='yes'
            ;;
        V)
            echo "$(basename $0) $VERSION"
            echo "Written by $AUTHOR"
            echo
            echo "$LICENSE"
            cleanup
            exit 0
            ;;
	E)
	    ENCODING=${OPTARG}
	    ;;
        \?)
            echo >&2 "Invalid option: -$OPTARG"
            cleanup
            exit 1
            ;;
        :)
            echo >&2 "Option -$OPTARG requires an argument."
            cleanup
            exit 1
    esac
done
shift $(($OPTIND - 1))

# Which printer
if [ "$METHOD" == "push" ]; then
    if [ -z $PRINTER ]; then
	echo >&2 "ERROR: No printer specified. Use '-P <printer>' or shell variable \$PRINTER"
	cleanup
	exit 1
    fi
elif [ "$METHOD" == "pull" ]; then
    available_queues=$(lpstat -s | awk '/pullprint.uio.no/ {print $4}' | cut -d/ -f4)
    num_queues=$(echo $available_queues | wc -w)

    # If there are no vendor queues available
    if [ $num_queues -eq 0 ]; then
	echo >&2 'ERROR: No pullprint vendor queues available on this host'
	cleanup
	exit 1
    fi

    # If there are only 1 vendor queue available
    if [ $num_queues -eq 1 ]; then
	PULLPRINT_QUEUE=$available_queues
    fi

    # Compare available queues to the one specified
    if [ $num_queues -gt 1 -a -n "$PULLPRINT_QUEUE" ]; then
	queue_exists=no
	for queue in $available_queues; do
	    [ "$queue" == "$PULLPRINT_QUEUE" ] && queue_exists=yes
	done
	if [ "$queue_exists" == "no" ]; then
	    echo >&2 "ERROR: Pullprint vendor queue '$PULLPRINT_QUEUE' doesn't exist"
	    echo >&2 -n '       Valid vendor queues on this host:'
	    for queue in $available_queues; do
		echo >&2 -n " $queue"
	    done
	    echo >&2
	    cleanup
	    exit 1
	fi
    fi

    # If vendor queue is still not decided
    if [ -z $PULLPRINT_QUEUE ]; then
	echo >&2 'ERROR: No pullprint vendor queue specified. Use "-P <queue>" or shell variable $PULLPRINT_QUEUE'
	echo >&2 -n '       Valid vendor queues on this host:'
	for queue in $available_queues; do
	    echo >&2 -n " $queue"
	done
	echo >&2
	cleanup
	exit 1
    fi
else
    echo >&2 "ERROR: Wrong printing method: '$METHOD'. This shouldn't happen."
    cleanup
    exit 1
fi		


#=====================================================================
# MAIN PROGRAM
#=====================================================================

# Files to print
input_files=("$@")

# Add stdin file if it exists
if [ -n "$stdin_file" -a -f "$stdin_file" ]; then
    input_files=("$stdin_file" "${input_files[@]}")
fi

# Error if there are no files to print
if [ ${#input_files[@]} -eq 0 ]; then
    echo >&2 'ERROR: No files to print.'
    exit 1
fi

# Get credentials if pushprint
if [ "$METHOD" == "push" ]; then
    if [ $KERBEROS == 'yes' ]; then
	kinit -R >/dev/null 2>&1 || kinit
	smbopts='-k'
    elif [ -n "${AUTH_FILE}" -a -r "${AUTH_FILE}" ]; then
	smbopts="-A ${AUTH_FILE}"
    else
	read -s -p "Password for ${PRINT_USER}: " PASSWORD
	echo
	smbopts="-N -U ${PRINT_USER}%${PASSWORD} -W ${WORKGROUP}"
    fi
fi

# Print files
for file in "${input_files[@]}"; do

    # Ignore bogus files
    if [ ! -e "$file" ]; then
        echo >&2 "WARNING: File '$file' doesn't exist, ignored"
        continue
    elif [ ! -f "$file" ]; then
        echo >&2 "WARNING: File '$file' isn't a regular file, ignored"
        continue
    elif [ ! -r "$file" ]; then
        echo >&2 "WARNING: File '$file' isn't readable, ignored"
        continue
    elif [ ! -s "$file" ]; then
        echo >&2 "WARNING: File '$file' has zero size, ignored"
        continue
    fi

    # Get type of file
    type=$(file -b "$file")

    # Convert file to PS if necessary
    print_file=
    converted='no'
    case "$type" in
        *'PDF document'*)
            converted='yes'
            print_file=$(mktemp --suff=.ps)
            pdftops -duplex "$file" $print_file
            ret=$?
            if [ $ret -ne 0 ]; then
                echo >&2 "WARNING: Couldn't convert '$file' to PostScript. Ignoring this file"
                [ -f $print_file ] && rm -f $print_file
                continue
            fi
            ;;
        *'PostScript document'*)
            print_file="$file"
            ;;
	*UTF-8*|'ASCII text')
	    converted='yes'
            print_file=$(mktemp --suff=.ps)
	    paps "$file" > $print_file
	    ;;
        *ISO-8859*)
	    converted='yes'
	    print_file=$(mktemp --suff=.ps)
	    paps --encoding=iso8859-1 "$file" > $print_file
	    ;;
        *Non-ISO*)
	    if [ -n "$ENCODING" ]; then
		converted='yes'
		print_file=$(mktemp --suff=.ps)
		paps --encoding="$ENCODING" "$file" > $print_file
		ret=$?
		if [  $ret -ne 0 ] ; then
		    echo >&2 "WARNING: Unknown encoding '$ENCODING', try 'iconv --list'. Skipping file '$file'"
		    rm -f $print_file
		    continue
		fi
	    else
                echo >&2 "WARNING: Unrecognized text file '$file', ignored. Try again with -E"
                rm -f $print_file
                continue
            fi
            ;;
        *)
            echo >&2 "WARNING: Ignoring file '$file': Unknown type: $type"
            continue
            ;;
    esac

    # Get proper title (for lpr)
    title=$(basename $file)

    # Print the (converted) file
    if [ "$METHOD" == "push" ]; then
	if [ $DRYRUN == "yes" ]; then
	    echo "(dryrun) smbclient //${SERVER}/${PRINTER} $smbopts -c \"${COMMAND}\" < \"${print_file}\""
	else
	    smbclient //${SERVER}/${PRINTER} $smbopts -c "${COMMAND}" < "${print_file}"
	fi
    elif [ "$METHOD" == "pull" ]; then
	if [ $DRYRUN == "yes" ]; then
	    echo "(dryrun) lpr -U ${PRINT_USER} -T \"$title\" -P pullprint_${PULLPRINT_QUEUE} \"${print_file}\""
	else
	    lpr -U ${PRINT_USER} -T "$title" -P pullprint_${PULLPRINT_QUEUE} "${print_file}"
	fi
    fi

    # Clean up temporary converted file
    [ $converted == 'yes' ] && rm -f $print_file
done

# Clean up and exit
cleanup
exit 0
